from .async_worker import AsyncWorker
from .event_loop import EventLoop
from .lazy_worker import LazyWorker
from .persistent_worker import PersistentWorker
from .single_worker import SingleWorker
from .worker import Worker
from .asyncio_mqtt import AsyncioMQTTHelper



from ..utils import Utils
from .attribute import Attribute

class Material(Attribute):
    """
    Material Attribute. For setting color and opacity.
    Usage: material=Material(...)
    """
    def __init__(self, **kwargs):
        if "opacity" in kwargs:
            # kwargs["transparent"] = True # need to be transparent to be opaque
            if kwargs["opacity"] > 1.0: # keep opacity between 0.0 and 1.0
                kwargs["opacity"] = float(kwargs["opacity"]) / 100
            kwargs["opacity"] = max(0.0, kwargs["opacity"])
            kwargs["opacity"] = min(kwargs["opacity"], 1.0)

        super().__init__(**kwargs)



from ..utils import *
from .attribute import Attribute
from .position import Position
from .rotation import Rotation
from .goto_url import GotoUrl
from .physics import Physics
from .scale import Scale
from .material import Material
from .color import Color

class Data(Attribute):
    """
    Data Attribute. Wraps all attributes in JSON.
    """
    def __init__(self, **kwargs):
        data = {}
        data = Data.update_data(data, kwargs)
        super().__init__(**data)

    @classmethod
    def update_data(cls, data, new_data):
        new_data = new_data.get("data", new_data)
        dash_words = []
        for k,v in new_data.items():
            # the dashes in these specific keys need to be replaced with underscores
            if k == "goto-url":
                dash_words += [k]
                k = "goto_url"
                if isinstance(v, dict):
                    data[k] = GotoUrl(**v)
                else:
                    data[k] = v

            # this could be called "clickable"
            if k == "click-listener":
                if "clickable" in data:
                    k = "clickable"
                else:
                    dash_words += [k]
                    k = "click_listener"
                data[k] = v

            # this could be called "physics"
            if k == "dynamic-body":
                if "physics" in data:
                    k = "physics"
                else:
                    dash_words += [k]
                    k = "dynamic_body"
                if isinstance(v, dict):
                    data[k] = Physics(**v)
                else:
                    data[k] = v

            # allow user to input tuples, lists, dicts, etc for specific Attributes.
            # everything gets converted to corresponding attribute
            if k == "position" and not isinstance(v, Position):
                if isinstance(v, (list,tuple)):
                    data[k] = Position(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Position(**v)
                else:
                    data[k] = v

            elif k == "rotation" and not isinstance(v, Rotation):
                if isinstance(v, (list,tuple)):
                    if len(v) == 3:
                        data[k] = Rotation(*v[:3], None)
                    else:
                        data[k] = Rotation(*v[:4])
                elif isinstance(v, dict):
                    data[k] = Rotation(**v)
                else:
                    data[k] = v

            elif k == "scale" and not isinstance(v, Scale):
                if isinstance(v, (list,tuple)):
                    data[k] = Scale(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Scale(**v)
                else:
                    data[k] = v

            elif k == "color" and not isinstance(v, Color):
                if isinstance(v, (list,tuple)):
                    color = Color(*v[:3])
                elif isinstance(v, dict):
                    color = Color(**v)
                elif isinstance(v, str):
                    color = Color(v)
                else:
                    color = v
                data[k] = color

            elif k == "material":
                if "color" in v:
                    color = v["color"]
                    if isinstance(color, (list,tuple)):
                        color = Color(*color[:3])
                    elif isinstance(color, dict):
                        color = Color(**color)
                    elif isinstance(color, str):
                        color = Color(color)
                    else:
                        color = v["color"]
                    v["color"] = color
                if isinstance(v, dict):
                    data[k] = Material(**v)
                else:
                    data[k] = v

            elif isinstance(v, Attribute):
                data[k] = v

            else:
                try:
                    # unknown attribute
                    data[k] = Attribute(**v)
                except:
                    data[k] = v

            # make False into None
            if (isinstance(v, bool) and v == False) or v is None:
                data[k] = None

        # delete elements with keys that have dashes
        for w in dash_words:
            if w in data:
                del data[w]

        return data



from .box import Box
from .camera import Camera
from .circle import Circle
from .cone import Cone
from .cylinder import Cylinder
from .dodecahedron import Dodecahedron
from .gltf import GLTF
from .icosahedron import Icosahedron
from .image import Image
from .landmarks import Landmarks
from .light import Light
from .line import Line
from .arena_object import Object
from .octahedron import Octahedron
from .particle import Particle
from .plane import Plane
from .ring import Ring
from .sphere import Sphere
from .tetrahedron import Tetrahedron
from .text import Text
from .thickline import ThickLine
from .torus import Torus
from .torus_knot import TorusKnot
from .triangle import Triangle

OBJECT_TYPE_MAP = {
    "box": Box,
    "cube": Box,
    "camera": Camera,
    "circle": Circle,
    "cone": Cone,
    "cylinder": Cylinder,
    "dodecahedron": Dodecahedron,
    "gltf-model": GLTF,
    "icosahedron": Icosahedron,
    "image": Image,
    "landmarks": Landmarks,
    "light": Light,
    "line": Line,
    "object": Object,
    "entity": Object,
    "octahedron": Octahedron,
    "particle": Particle,
    "plane": Plane,
    "ring": Ring,
    "sphere": Sphere,
    "tetrahedron": Tetrahedron,
    "text": Text,
    "thickline": ThickLine,
    "torus": Torus,
    "torusKnot": TorusKnot,
    "triangle": Triangle,
}



import math
from ..utils import Utils
from .attribute import Attribute

class Position(Attribute):
    """
    Position Attribute.
    Usage: position=Position(x,y,z)
    """
    def __init__(self, x=0, y=0, z=0):
        super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z))

    def to_str(self):
        return Utils.tuple_to_string((self.x, self.y, self.z))

    def distance_to(self, pos):
        return math.sqrt((self.x-pos.x)**2 + (self.y-pos.y)**2 + (self.z-pos.z)**2)



from ..attributes import Attribute
from .arena_object import Object

class Landmarks(Object):
    """Landmarks class. Holds a list of landmarks."""
    def __init__(self):
        super().__init__(object_id="scene-landmarks", persist=True)
        self.type = "landmarks" # force type to be "landmarks"
        self.data.landmarks = []

    def add(self, obj, label):
        if isinstance(obj, Attribute):
            object_id = obj.object_id
        else:
            object_id = obj
        self.data.landmarks += [{
            "object_id": object_id,
            "label": label
        }]



"""
ARENA-py

Draw objects and run programs in the ARENA using Python!
"""

__version__     = "0.1.25"
__author__      = "CONIX Research Center"

__license__     = "BSD 3-Clause"
__maintainer__  = "WiSE Lab, Carnegie Mellon University"

from .scene import *



import re
import webcolors
from .attribute import Attribute

class Color(Attribute):
    """
    Color Attribute.
    Usage: color=Color(red,green,blue)
    """
    def __init__(self, red=0, green=0, blue=0):
        if isinstance(red, str):
            # hex to tuple to Color
            color = red.lstrip("#")
            hexcolor = re.search(r"^(?:[0-9a-fA-F]{3}){1,2}$", color)
            if not hexcolor:
                try:
                    wcrgb = webcolors.name_to_rgb(color)
                except:
                    wcrgb = webcolors.hex_to_rgb("#0"+color)
                c = (wcrgb.red, wcrgb.green, wcrgb.blue)
            else:
                c = tuple(int(color[c:c+2], 16) for c in (0, 2, 4))
            red, blue, green = c
        super().__init__(red=red, green=green, blue=blue)

    @property
    def hex(self):
        return "#{:02x}{:02x}{:02x}".format(self.red, self.green, self.blue)



from ..utils import Utils
from .attribute import Attribute

class Morph(Attribute):
    """
    Morph Attribute.
    Usage: Morph(...)
    """
    def __init__(self, morphtarget, value):
        self.morphtarget = str(morphtarget)
        self.value = str(value)



from .arena_object import Object

class Ring(Object):
    """
    Class for Ring in the ARENA.
    """
    object_type = "ring"

    def __init__(self, **kwargs):
        super().__init__(object_type=Ring.object_type, **kwargs)



from ..base_object import *
from ..attributes import Animation, Data
from ..utils import *
import uuid


class Object(BaseObject):
    """
    Object class. Defines a generic object in the ARENA.
    """

    type = "object"
    object_type = "entity"
    all_objects = {} # dict of all objects created so far

    def __init__(self, evt_handler=None, update_handler=None, **kwargs):
        # "object_id" is required in kwargs, defaulted to random uuid4
        object_id = kwargs.get("object_id", str(uuid.uuid4()))
        if "object_id" in kwargs: del kwargs["object_id"]

        # "persist" is required in kwargs, defaulted to false
        persist = kwargs.get("persist", False)
        if "persist" in kwargs: del kwargs["persist"]

        # special case for "parent" (can be an Object)
        if "parent" in kwargs and isinstance(kwargs["parent"], Object):
            kwargs["parent"] = kwargs["parent"].object_id

        # "ttl" is optional
        ttl = kwargs.get("ttl", None)
        if "ttl" in kwargs: del kwargs["ttl"]

        # remove timestamp, if exists
        if "timestamp" in kwargs: del kwargs["timestamp"]

        # remove "updatedAt", if exists
        if "updatedAt" in kwargs: del kwargs["updatedAt"]

        # remove "action", if exists
        if "action" in kwargs: del kwargs["action"]

        # default "object_type" to entity
        if "object_type" not in kwargs:
            kwargs["object_type"] = Object.object_type

        # print warning if object is being created with the same id as an existing object
        if Object.exists(object_id):
            if not Object.get(object_id).persist:
                print("[WARNING]", f"An object with object_id of {object_id} was already created. The previous object will be overwritten.")
            Object.remove(Object.get(object_id))

        # setup attributes in the "data" field
        data = kwargs.get("data", kwargs)
        data = Data(**data)
        super().__init__(
                object_id=object_id,
                type=Object.type,
                persist=persist,
                data=data
            )
        if ttl:
            self.ttl = ttl

        self.evt_handler = evt_handler
        self.update_handler = update_handler
        self.animations = []

        # add current object to all_objects dict
        Object.add(self)

    def update_attributes(self, evt_handler=None, update_handler=None, **kwargs):
        if evt_handler:
            self.evt_handler = evt_handler

        if update_handler:
            self.update_handler = update_handler

        if "data" not in self:
            return

        # update "persist", and "ttl"
        self.persist = kwargs.get("persist", self.persist)
        if "ttl" in self:
            self.ttl = kwargs.get("ttl", self.ttl)

        data = self.data
        Data.update_data(data, kwargs)

        if self.update_handler:
            self.update_handler(self)

    def dispatch_animation(self, animation):
        if isinstance(animation, (tuple, list)):
            self.animations += list(animation)
        elif isinstance(animation, Animation):
            self.animations += [animation]
        return self.animations

    def remove_animation_at_index(self, idx):
        if 0 <= idx < len(self.animations):
            return self.animations.pop(idx)
        return -1

    def clear_animations(self):
        self.animations = []

    @property
    def clickable(self):
        return "click_listener" in self.data or "clickable" in self.data

    def json_preprocess(self, **kwargs):
        # kwargs are for additional param to add to json, like "action":"create"
        json_payload = { k:v for k,v in vars(self).items() if k != "evt_handler" and \
                                            k != "update_handler" and k != "animations" }
        json_payload.update(kwargs)
        return json_payload

    def json_postprocess(self, json_payload, json_data): # to be done by subclasses, if needed
        pass

    def json(self, **kwargs):
        json_data = {}
        json_payload = self.json_preprocess(**kwargs)
        data = vars(json_payload["data"])

        for k,v in data.items():
            if v is None:
                json_data[k] = v

            # color should be a hex string
            elif "color" == k:
                json_data["color"] = v.hex

            elif "material" == k:
                json_data["material"] = vars(v).copy()
                if "color" in v:
                    color = v["color"]
                    json_data["material"]["color"] = color.hex

            # rotation should be in quaternions
            elif "rotation" == k:
                rot = data["rotation"]
                if rot.is_quaternion:
                    json_data["rotation"] = rot
                else:
                    # remove w if euler
                    rot_json = vars(rot).copy()
                    del rot_json["w"]
                    json_data["rotation"] = rot_json

            # handle special case where "physics" should be "dynamic-body"
            elif "physics" == k or "dynamic_body" == k:
                json_data["dynamic-body"] = v

            # handle special case where "clickable" should be "click-listener"
            elif "clickable" == k or "click_listener" == k:
                json_data["click-listener"] = v

            # remove underscores from specific keys
            elif "goto_url" == k:
                json_data["goto-url"] = v

            elif "collision_listener" == k:
                json_data["collision-listener"] = v

            elif "animation_mixer" == k:
                json_data["animation-mixer"] = v

            # for animation, replace "start" and "end" with "from" and "to"
            elif isinstance(k, str) and "animation" == k[:len("animation")]:
                animation = vars(v).copy()
                Utils.dict_key_replace(animation, "start", "from")
                Utils.dict_key_replace(animation, "end", "to")
                json_data[k] = animation
            else:
                json_data[k] = v

        json_payload["data"] = json_data
        self.json_postprocess(json_payload, json_data)
        return self.json_encode(json_payload)

    # methods for global objects dictionary
    @classmethod
    def get(cls, object_id):
        return Object.all_objects.get(object_id, None)

    @classmethod
    def add(cls, obj):
        object_id = obj.object_id
        Object.all_objects[object_id] = obj

    @classmethod
    def remove(cls, obj):
        object_id = obj.object_id
        del Object.all_objects[object_id]

    @classmethod
    def exists(cls, object_id):
        return object_id in Object.all_objects



import json
from .arena_object import Object
from ..attributes import Morph

class GLTF(Object):
    """
    Class for GLTF Models in the ARENA.
    """
    object_type = "gltf-model"

    def __init__(self, url="", **kwargs):
        self.morphs = {}
        super().__init__(object_type=GLTF.object_type, url=url, **kwargs)

    def update_morph(self, morph):
        if isinstance(morph, (list,tuple)):
            for m in morph:
                self.morphs[m.morphtarget] = m
        elif isinstance(morph, Morph):
            self.morphs[morph.morphtarget] = morph
        return self.morphs

    def remove_morph(self, morph):
        if morph.morphtarget in self.morphs:
            del self.morphs[morph.morphtarget]

    def clear_morphs(self):
        self.morphs = {}

    def json_preprocess(self, **kwargs):
        # kwargs are for additional param to add to json, like "action":"create"
        json_payload = { k:v for k,v in vars(self).items() if not callable(v) and k != "animations" and k != "morphs" }
        json_payload.update(kwargs)
        return json_payload

    def json_postprocess(self, json_payload, json_data):
        for i,morph in enumerate(self.morphs.values()):
            json_data[f"gltf-morph__{i}"] = vars(morph)

class Model(GLTF):
    """
    Another name for GLTF.
    """



import os
import sys
import json
import random
import socket
import asyncio

import paho.mqtt.client as mqtt
from datetime import datetime
from inspect import signature

from .attributes import *
from .objects import *
from .events import *
from .utils import *
from .event_loop import *

from . import auth

class Scene(object):
    """
    Gives access to an ARENA scene.
    Wrapper around Paho MQTT client and EventLoop.
    Can create and execute various user-defined functions/tasks.
    """
    def __init__(
                self,
                network_latency_interval = 10000,  # run network latency update every 10s
                on_msg_callback = None,
                new_obj_callback = None,
                user_join_callback = None,
                user_left_callback = None,
                delete_obj_callback = None,
                end_program_callback = None,
                debug = False,
                **kwargs
            ):
        if os.environ.get("MQTTH"):
            self.host  = os.environ["MQTTH"]
        elif "host" in kwargs:
            self.host = kwargs["host"]
            print("Cannot find MQTTH environmental variable, using input parameter instead.")
        else:
            sys.exit("mqtt host argument (host) is unspecified, aborting...")

        if os.environ.get("REALM"):
            self.realm  = os.environ["REALM"]
        elif "realm" in kwargs:
            self.realm = kwargs["realm"]
            print("Cannot find REALM environmental variable, using input parameter instead.")
        else:
            sys.exit("realm argument (realm) is unspecified, aborting...")

        if os.environ.get("SCENE"):
            self.scene  = os.environ["SCENE"]
        elif "scene" in kwargs:
            self.scene = kwargs["scene"]
            print("Cannot find SCENE environmental variable, using input parameter instead.")
        else:
            sys.exit("scene argument (scene) is unspecified, aborting...")

        self.debug = debug

        print("=====")
        # do user auth
        username = None
        password = None
        if os.environ.get("ARENA_USERNAME") and os.environ.get("ARENA_PASSWORD"):
            username = os.environ["ARENA_USERNAME"]
            password = os.environ["ARENA_PASSWORD"]
        else:
            username = auth.authenticate_user(self.host, debug=self.debug)

        if os.environ.get("NAMESPACE"):
            self.namespace = os.environ["NAMESPACE"]
        elif "namespace" not in kwargs:
            self.namespace = username
        else:
            self.namespace = kwargs["namespace"]

        self.mqttc_id = "pyClient-" + self.generate_client_id()

        # set up scene variables
        self.namespaced_scene =  f"{self.namespace}/{self.scene}"

        self.root_topic = f"{self.realm}/s/{self.namespaced_scene}"
        self.scene_topic = f"{self.root_topic}/#"   # main topic for entire scene
        self.persist_url = f"https://{self.host}/persist/{self.namespaced_scene}"

        self.latency_topic = "$NETWORK/latency"     # network graph latency update
        self.ignore_topic = f"{self.root_topic}/{self.mqttc_id}/#" # ignore own messages

        self.mqttc = mqtt.Client(
            self.mqttc_id, clean_session=True
        )

        # do scene auth
        if username is None or password is None:
            data = auth.authenticate_scene(
                            self.host, self.realm,
                            self.namespaced_scene, username,
                            self.debug
                        )
            if 'username' in data and 'token' in data:
                username = data["username"]
                password = data["token"]
        self.mqttc.username_pw_set(username=username, password=password)
        print("=====")

        # set up callbacks
        self.on_msg_callback = on_msg_callback
        self.new_obj_callback = new_obj_callback
        self.delete_obj_callback = delete_obj_callback
        self.user_join_callback = user_join_callback
        self.user_left_callback = user_left_callback
        self.end_program_callback = end_program_callback

        self.unspecified_object_ids = set() # objects that exist in the scene,
                                          # but this scene instance does not
                                          # have a reference to
        self.users = {} # dict of all users
        self.landmarks = Landmarks() # scene landmarks

        self.task_manager = EventLoop(self.disconnect)

        aioh = AsyncioMQTTHelper(self.task_manager, self.mqttc)

        # have all tasks wait until mqtt client is connected before starting
        self.mqtt_connect_evt = asyncio.Event()
        self.mqtt_connect_evt.clear()

        # set paho mqtt callbacks
        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect

        # add main message processing + callbacks loop to tasks
        self.run_async(self.process_message)

        # update network latency every network_latency_interval secs
        self.run_forever(self.network_latency_update,
                         interval_ms=network_latency_interval)

        self.msg_queue = asyncio.Queue()

        # connect to mqtt broker
        if "port" in kwargs:
            self.mqttc.connect(self.host, kwargs["port"])
        else:
            self.mqttc.connect(self.host)
        self.mqttc.socket().setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)

        print(f"Loading: https://{self.host}/{self.namespace}/{self.scene}, realm={self.realm}")

    def generate_client_id(self):
        """Returns a random 6 digit id"""
        return str(random.randrange(100000, 999999))

    def network_latency_update(self):
        """Update client latency in $NETWORK/latency"""
        # publish empty message with QoS of 2 to update latency
        self.mqttc.publish(self.latency_topic, "", qos=2)

    def run_once(self, func=None, **kwargs):
        """Runs a user-defined function on startup"""
        if func is not None:
            w = SingleWorker(func, self.mqtt_connect_evt, **kwargs)
            self.task_manager.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_once(func):
                self.run_once(func, **kwargs)
                return func
            return _run_once

    def run_after_interval(self, func=None, interval_ms=1000, **kwargs):
        """Runs a user-defined function after a interval_ms milliseconds"""
        if func is not None:
            if interval_ms < 0:
                print("Invalid interval! Defaulting to 1000ms")
                interval_ms = 1000
            w = LazyWorker(func, self.mqtt_connect_evt, interval_ms, **kwargs)
            self.task_manager.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_after_interval(func):
                self.run_after_interval(func, interval_ms, **kwargs)
                return func
            return _run_after_interval

    def run_async(self, func=None, **kwargs):
        """Runs a user defined aynscio function"""
        if func is not None:
            w = AsyncWorker(func, self.mqtt_connect_evt, **kwargs)
            self.task_manager.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_async(func):
                self.run_async(func, **kwargs)
                return func
            return _run_async

    def run_forever(self, func=None, interval_ms=1000, **kwargs):
        """Runs a function every interval_ms milliseconds"""
        if func is not None:
            if interval_ms < 0:
                print("Invalid interval! Defaulting to 1000ms")
                interval_ms = 1000
            t = PersistentWorker(func, self.mqtt_connect_evt, interval_ms, **kwargs)
            self.task_manager.add_task(t)
        else:
            # if there is no func, we are in a decorator
            def _run_forever(func):
                self.run_forever(func, interval_ms, **kwargs)
                return func
            return _run_forever

    def run_tasks(self):
        """Run event loop"""
        print("Connecting to the ARENA...")
        self.task_manager.run()

    def stop_tasks(self):
        """Stop event loop"""
        self.task_manager.stop()

    async def sleep(self, interval_ms):
        """Public function for sleeping in async functions"""
        await asyncio.sleep(interval_ms / 1000)

    def on_connect(self, client, userdata, flags, rc):
        """Paho MQTT client on_connect callback"""
        if rc == 0:
            self.mqtt_connect_evt.set()

            # listen to all messages in scene
            client.subscribe(self.scene_topic)
            client.message_callback_add(self.scene_topic, self.on_message)

            # create ARENA-py Objects from persist server
            # no need to return anything here
            self.get_persisted_objs()

            print("Connected!")
            print("=====")
        else:
            print(f"Connection error! Result code: {rc}")

    def on_message(self, client, userdata, msg):
        # ignore own messages
        if mqtt.topic_matches_sub(self.ignore_topic, msg.topic):
            return

        self.msg_queue.put_nowait(msg)

    async def process_message(self):
        """Main message processing function"""
        while True:
            msg = await self.msg_queue.get()

            # extract payload
            try:
                payload_str = msg.payload.decode("utf-8", "ignore")
                payload = json.loads(payload_str)
            except:
                return

            # update object attributes, if possible
            if "object_id" in payload:
                # parese payload
                object_id = payload.get("object_id", None)
                action = payload.get("action", None)

                data = payload.get("data", {})
                object_type = data.get("object_type", None)

                event = None

                # create/get object from object_id
                if object_id in self.all_objects:
                    obj = self.all_objects[object_id]
                else:
                    ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                    obj = ObjClass(**payload)

                # react to action accordingly
                if action:
                    if action == "clientEvent":
                        event = Event(**payload)
                        if obj.evt_handler:
                            self.callback_wrapper(obj.evt_handler, event, payload)
                            continue

                    elif action == "delete":
                        if Camera.object_type in object_id: # object is a camera
                            if object_id in self.users and self.user_left_callback:
                                self.callback_wrapper(
                                        self.user_left_callback,
                                        self.users[object_id],
                                        payload
                                    )
                        elif self.delete_obj_callback:
                            self.callback_wrapper(self.delete_obj_callback, obj, payload)
                        continue

                    else: # create/update
                        obj.update_attributes(**payload)

                # call new message callback for all messages
                if self.on_msg_callback:
                    if not event:
                        self.callback_wrapper(self.on_msg_callback, obj, payload)
                    else:
                        self.callback_wrapper(self.on_msg_callback, event, payload)

                # run user_join_callback when user is found
                if object_type and object_type == Camera.object_type:
                    if object_id not in self.users:
                        if object_id in self.all_objects:
                            self.users[object_id] = obj
                        else:
                            self.users[object_id] = Camera(**payload)

                        if self.user_join_callback:
                            self.callback_wrapper(
                                    self.user_join_callback,
                                    self.users[object_id],
                                    payload
                                )

                # if its an object the library has not seen before, call new object callback
                elif object_id not in self.unspecified_object_ids and self.new_obj_callback:
                    self.callback_wrapper(self.new_obj_callback, obj, payload)
                    self.unspecified_object_ids.add(object_id)

    def callback_wrapper(self, func, arg, msg):
        """Checks for number of arguments for callback"""
        if len(signature(func).parameters) != 3:
            print("[DEPRECATED]", "Callbacks and handlers now take 3 arguments: (scene, obj/evt, msg)!")
            func(arg)
        else:
            func(self, arg, msg)

    def on_disconnect(self, client, userdata, rc):
        """Paho MQTT client on_disconnect callback"""
        if rc == 0:
            print("Disconnected from the ARENA!")
        else:
            print(f"Disconnect error! Result code: {rc}")

    def disconnect(self):
        """Disconnects Paho MQTT client"""
        if self.end_program_callback:
            self.end_program_callback(self)
        self.mqttc.disconnect()

    def generate_custom_event(self, evt, action="clientEvent"):
        """Publishes an custom event. Could be user or library defined"""
        return self._publish(evt, action)

    def generate_click_event(self, obj, type="mousedown", **kwargs):
        """Publishes an click event"""
        _type = type
        evt = Event(object_id=obj.object_id,
                    type=_type,
                    position=obj.data.position,
                    source=self.mqttc_id,
                    **kwargs)
        return self.generate_custom_event(evt, action="clientEvent")

    def manipulate_camera(self, cam, **kwargs):
        """Publishes a camera manipulation event"""
        if kwargs.get("position"):
            if isinstance(kwargs["position"], (list, tuple)):
                kwargs["position"] = Position(*kwargs["position"])
            elif isinstance(kwargs["position"], dict):
                kwargs["position"] = Position(**kwargs["position"])

        if kwargs.get("rotation"):
            if isinstance(kwargs["rotation"], (list, tuple)):
                kwargs["rotation"] = Rotation(*kwargs["rotation"])
            elif isinstance(kwargs["rotation"], dict):
                kwargs["rotation"] = Rotation(**kwargs["rotation"])

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id,
                    type="camera-override",
                    object_type=Camera.object_type,
                    **kwargs)
        return self.generate_custom_event(evt, action="update")

    def look_at(self, cam, target):
        """Publishes a camera manipulation event"""
        if isinstance(target, tuple) or isinstance(target, list):
            target = Position(*target)
        elif isinstance(target, dict):
            target = Position(**target)
        elif isinstance(target, Object):
            target = target.object_id

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id,
                    type="camera-override",
                    object_type="look-at",
                    target=target)
        return self.generate_custom_event(evt, action="update")

    def add_landmark(self, obj, label):
        """Adds a landmark to the scene"""
        if isinstance(obj, Object):
            landmark_id = obj.object_id
        else:
            landmark_id = Object
        # object must be persisted to make landmarks make sense
        obj.persist = True
        self.add_object(obj)
        self.landmarks.add(landmark_id, label)
        return self._publish(self.landmarks, "create")

    @property
    def all_objects(self):
        """Returns all the objects in a scene"""
        return Object.all_objects

    def add_object(self, obj):
        """Public function to create an object"""
        res = self._publish(obj, "create")
        self.run_animations(obj)
        return res

    def add_objects(self, objs):
        """Public function to create multiple objects in a list"""
        for obj in objs:
            self.add_objects(obj)
        return len(objs)

    def update_object(self, obj, **kwargs):
        """Public function to update an object"""
        if kwargs:
            obj.update_attributes(**kwargs)
        res = self._publish(obj, "update")
        self.run_animations(obj)
        return res

    def update_objects(self, objs, **kwargs):
        """Public function to update multiple objects in a list"""
        for obj in objs:
            self.update_object(obj, **kwargs)
        return len(objs)

    def delete_object(self, obj):
        """Public function to delete an object"""
        payload = {
            "object_id": obj.object_id
        }
        Object.remove(obj)
        return self._publish(payload, "delete")

    def run_animations(self, obj):
        """Runs all dispatched animations"""
        if obj.animations:
            payload = {
                "object_id": obj.object_id,
                "type": obj.type,
                "data": {"object_type": obj.data.object_type}
            }
            for i,animation in enumerate(obj.animations):
                if isinstance(animation, AnimationMixer):
                    payload["data"][f"animation-mixer"] = vars(animation)
                else:
                    anim = vars(animation).copy()
                    if i == 0:
                        payload["data"][f"animation"] = anim
                    else:
                        payload["data"][f"animation__{i}"] = anim
                    Utils.dict_key_replace(anim, "start", "from")
                    Utils.dict_key_replace(anim, "end", "to")
            obj.clear_animations()
            return self._publish(payload, "dispatch_animation")

    def _publish(self, obj, action):
        """Publishes to mqtt broker with "action":action"""
        topic = f"{self.root_topic}/{self.mqttc_id}/{obj['object_id']}"
        d = datetime.now().isoformat()[:-3]+"Z"

        if action == "delete":
            payload = obj
            payload["action"] = "delete"
            payload["timestamp"] = d
            payload = json.dumps(payload)
        elif action == "dispatch_animation":
            payload = obj
            payload["action"] = "update"
            payload["timestamp"] = d
            payload = json.dumps(payload)
        else:
            payload = obj.json(action=action, timestamp=d)

        self.mqttc.publish(topic, payload, qos=0)
        if self.debug: print("[publish]", topic, payload)
        return payload

    def get_persisted_obj(self, object_id):
        """Returns a dictionary for a persisted object."""
        obj = None
        if object_id in self.all_objects:
            obj = self.all_objects[object_id]
            obj.persist = True
        else:
            # pass token to persist
            data = auth.urlopen(url=f"{self.persist_url}/{object_id}", creds=True)
            output = json.loads(data)
            if len(output) > 0:
                output = output[0]

                object_id = output["object_id"]
                data = output["attributes"]
                object_type = data.get("object_type", None)

                ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                obj = ObjClass(object_id=object_id, data=data)
                obj.persist = True

        return obj

    def get_persisted_objs(self):
        """Returns a dictionary of persisted objects. [TODO] check object_type"""
        objs = {}
        # pass token to persist
        data = auth.urlopen(url=self.persist_url, creds=True)
        output = json.loads(data)
        for obj in output:
            if obj["type"] == Object.object_type or obj["type"] == Object.type:
                object_id = obj["object_id"]
                data = obj["attributes"]

                if object_id in self.all_objects:
                    persisted_obj = self.all_objects[object_id]
                    persisted_obj.persist = True
                else:
                    object_type = data.get("object_type", None)
                    ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                    persisted_obj = ObjClass(object_id=object_id, data=data)
                    persisted_obj.persist = True

                objs[object_id] = persisted_obj

        return objs

    def get_persisted_scene_option(self):
        """Returns a dictionary for scene-options. [TODO] wrap the output as a BaseObject"""
        scene_opts_url = f"{self.persist_url}?type=scene-options"
        # pass token to persist
        data = auth.urlopen(url=scene_opts_url, creds=True )
        output = json.loads(data)
        return output


    def get_writable_scenes(self):
        """ Request list of scene names for logged in user account that user has publish permission for.
        Returns: list of scenes.
        """
        return auth.get_writable_scenes(host=self.host, debug=self.debug)


    def message_callback_add(self, sub, callback):
        """Subscribes to new topic and adds callback"""
        self.mqttc.message_callback_add(sub, callback)

    def message_callback_remove(self, sub):
        """Unsubscribes to topic and removes filter for callback"""
        self.mqttc.message_callback_remove(sub)

    def get_user_list(self):
        """Returns a list of users"""
        return self.users.values()

class Arena(Scene):
    """
    Another name for Scene.
    """



from .arena_object import Object
from ..attributes import Position

class Line(Object):
    """
    Class for Line in the ARENA.
    """
    object_type = "line"

    def __init__(self, start=Position(0,0,0), end=Position(10,10,10), **kwargs):
        super().__init__(object_type=Line.object_type, start=start, end=end, **kwargs)




import os
import signal
import asyncio

class EventLoop(object):
    """
    Wrapper for an asyncio event loop.
    """
    def __init__(self, shutdown_func=None):
        self.tasks = []
        self.loop = asyncio.get_event_loop()
        if os.name == 'nt': # Windows doesnt have SIGHUP signal
            self.signals = (signal.SIGTERM, signal.SIGINT)
        else:
            self.signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        self.shutdown_func = shutdown_func

    async def _shutdown(self):
        if self.shutdown_func:
            self.shutdown_func()
        self.future.cancel()
        self.loop.stop()

    def add_task(self, worker):
        self.tasks += [worker.run()]

    def shutdown_wrapper(self, *args):
        asyncio.ensure_future(self._shutdown())

    def create_future(self):
        return self.loop.create_future()

    def run(self):
        # cancellation of Future ensures all tasks (even if not completed) are cancelled
        self.future = asyncio.gather(*self.tasks)

        # register signals
        for s in self.signals:
            if os.name == 'nt': # Windows does not add_signal_handler implemented
                signal.signal(s, self.shutdown_wrapper)
            else:
                self.loop.add_signal_handler(
                    s, self.shutdown_wrapper
                )

        # run event loop
        try:
            self.loop.run_forever()
        except asyncio.exceptions.CancelledError:
            pass
        finally:
            self.loop.close()

    def stop(self):
        self.loop.stop()



from .arena_object import Object

class Light(Object):
    """
    Class for Light in the ARENA.
    """
    object_type = "light"

    def __init__(self, **kwargs):
        super().__init__(object_type=Light.object_type, **kwargs)



import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="arena-py",
    version="0.1.25",
    author="Conix Research Center",
    author_email="info@conix.io",
    license="BSD 3-clause \"New\" or \"Revised License\"",
    description="Draw objects and run programs in the ARENA using Python!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/conix-center/ARENA-py",
    packages=setuptools.find_packages(),
    install_requires=[
        "aiohttp>=3.7.4",
        "paho-mqtt~=1.5.0",
        "numpy~=1.19.5",
        "requests~=2.23.0",
        "scipy~=1.5.4",
        "webcolors~=1.3",
        "google_auth_oauthlib~=0.4.2",
        "PyJWT~=2.0.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from .arena_object import Object

class TorusKnot(Object):
    """
    Class for TorusKnot in the ARENA.
    """
    object_type = "torusKnot"

    def __init__(self, **kwargs):
        super().__init__(object_type=TorusKnot.object_type, **kwargs)



from .arena_object import Object

class Circle(Object):
    """
    Class for Circle in the ARENA.
    """
    object_type = "circle"

    def __init__(self, **kwargs):
        super().__init__(object_type=Circle.object_type, **kwargs)
